"""
Base class for block provider implementations
"""
import json
import os

from . import log

class BlockObj(object):
    """
    Class representing a path to a block device.

    Add properties to avoid direct access
    """

    def __init__(self, path, serial=None):
        self._path = path
        self._serial = serial
        self._size = 0

    @property
    def provider(self):
        raise NotImplementedError()

    def get_path(self):
        """get the path"""
        return self._path

    def set_serial(self, serial):
        self._serial = serial

    def persist(self):
        """Notify provider that this path will be persistent"""

        # Set the block scheduler
        device = os.path.realpath(self._path)[5:]
        scheduler_path = '/sys/block/{}/queue/scheduler'.format(device)
        if os.path.exists(scheduler_path):
            with open(scheduler_path, 'w') as scheduler:
                log.sm_log('Setting noop scheduler for {}'.format(device))
                try:
                    scheduler.write("{}\n".format('noop'))
                except OSError:
                    pass

    def get_device_info(self):
        """
        Get any device specific information to present to the user
        """
        return {}

    def close(self):
        pass

    def __str__(self):
        """ Get a human readable string """
        return "Device: provider {0}, path {1}".format(self.provider, self._path)


class BlockList(object):
    """
    Class representing a list of BlockObj

    Add properties to avoid direct access
    """

    def __init__(self):
        self._devs = []

    def __iter__(self):
        return iter(self._devs)

    def add(self, blk_obj):
        """Add obj"""
        self._devs.append(blk_obj)

    def extend(self, blk_list):
        self._devs.extend(blk_list)


class BlockProviderException(Exception):
    """
    General exception type for block provider errors

    Message should be formatted JSON response.
    """
    @staticmethod
    def format_json_options(option_key, options):
        """
        Format a set of additional transport parameters into JSON
        """
        result = {'options': {option_key: options}}

        return json.dumps(result, ensure_ascii=True)

    @staticmethod
    def format_error_message(message):
        """
        Format an error message into a JSON reponse
        """
        return json.dumps({'error': message})


class BlockProvider(object):
    """
    Abstract block provider base class
    """

    def __init__(self):
        raise NotImplementedError("Cannot create an instance")
