"""
Transport handler dispatcher. Selects the correct transport provider to
dispatch to and executes the call.
"""
import json

from .hba import HBA
from .iscsi import ISCSI
from .scsiid_mapper import ScsiIdMapper as Scsi_mapper
from .blockprovider import BlockProviderException
from .multipath import MultiObj

HANDLERS = {}


class DispatcherException(Exception):
    """
    General exception type for errors in dispatching requests
    """
    pass


class Dispatcher(object):
    """
    Request dispatcher for block provider requests
    """
    @staticmethod
    def create_device_info_list(dev_list):
        def device_info(device):
            info = {
                'path': device.path,
                'size': str(device.size)
            }

            info.update(device.get_device_info())

            return info
        return [{'value': x.scsi_id, 'info': device_info(x)} for x in dev_list]

    @staticmethod
    def dispatch(configuration, rescan):
        """
        Dispatch the request to the block provider.
        Configuration must contain a "provider" key to allow for dispatch

        Returns:
        Instantiated BlockProvider object
        """
        provider = configuration["provider"]
        if provider not in HANDLERS:
            raise DispatcherException("Unsupported handler")

        # Get the devices from the block provider
        blk_list = HANDLERS[provider].get_paths(configuration, rescan)
        # Get the devices organized by scsi_id
        dev_list = Scsi_mapper.get_devs(blk_list)
        # Return the wanted device
        if 'ScsiId' not in configuration:
            # no SCSI id so list the ones we have
            raise BlockProviderException(
                BlockProviderException.format_json_options(
                    'ScsiId',
                    Dispatcher.create_device_info_list(dev_list)))

        scsi_obj = dev_list.get_obj(configuration['ScsiId'])

        if not scsi_obj:
            raise BlockProviderException(
                BlockProviderException.format_json_options(
                    'ScsiId',
                    Dispatcher.create_device_info_list(dev_list)))

        # Multipath section
        if ('multipath' in configuration
                and configuration['multipath'] == 'yes'):
            return MultiObj(scsi_obj)

        return scsi_obj

    @staticmethod
    def supported(provider_key):
        """
        Return the name of the supported handlers
        """
        return provider_key in [x for x in HANDLERS]


    @staticmethod
    def register(handler_class):
        """
        Register a block provider with the dispatcher
        """
        # TODO: If we need to log this then we should have a logger
        #print "Registering {}".format(handler_class)
        HANDLERS[handler_class.provider()] = handler_class

# Don't really like this, would be nicer if things could "auto" register
Dispatcher.register(ISCSI)
Dispatcher.register(HBA)
