"""
Implementation of BlockProvider to handle HBA targets
"""
import errno
import glob
import os
import re
import time

from .blockprovider import BlockProvider, BlockProviderException
from .blockprovider import BlockObj


FC_TRANSPORT_PATH = '/sys/class/fc_transport/'
SAS_DEVICE_FNMATCH = ('/sys/class/sas_end_device/end_device-*'
                      '/device/target*/*/block/*')
FC_HOST_FNMATCH = ('/sys/class/fc_host/*/device/scsi_host/*/scan')


class HBA(BlockProvider):
    """
    Class providing paths to HBA connected devices.
    """

    @staticmethod
    def provider():
        """
        Return the provider key for selecting this transport
        """
        return 'hba'

    @classmethod
    def make_blk_obj(cls, adapter, channel, target_id, lun, dev):
        """
        Create a block definition object
        """
        return HBABlockObj('/dev/{}'.format(dev),
                           adapter, channel, target_id, lun, cls.provider())

    @staticmethod
    def _get_fibre_channel_targets():
        """
        List the fibre channel targets known by the system
        """
        targets = []
        try:
            targets = os.listdir(FC_TRANSPORT_PATH)
        except OSError as oserror:
            if oserror.errno != errno.ENOENT:
                raise
            # No FC targets - FC_TRANSPORT_PATH does not exist. pass.
        return targets

    @staticmethod
    def _make_fibre_channel_device_path(target):
        return os.path.join(os.path.join(FC_TRANSPORT_PATH, target), 'device')

    @staticmethod
    def _find_fibre_channel_device_link(path, device):
        block_path = os.path.join(path, device, 'block')
        if os.path.exists(block_path):
            return os.listdir(block_path)[0]

        scsi_generic_path = os.path.join(path, device, 'scsi_generic')
        if os.path.exists(scsi_generic_path):
            return os.listdir(scsi_generic_path)[0]

        return ""

    @staticmethod
    def _get_fibre_channel_devices(targets):
        """
        List the devices for the supplied fibre channel targets
        """
        devs = []

        for target in targets:
            key = target.replace('target', '')
            target_path = HBA._make_fibre_channel_device_path(target)
            try:
                listed_devices = os.listdir(target_path)
                for device in [x for x in listed_devices if x.startswith(key)]:
                    (adapter, channel, target_id, lun) = device.split(':')
                    devs.append(
                        (adapter, channel, target_id, lun,
                         HBA._find_fibre_channel_device_link(target_path,
                                                             device)))
            except OSError as oserror:
                if oserror.errno != errno.ENOENT:
                    raise
                # target_path does not exist. pass.
        return devs

    @staticmethod
    def get_sas_devices():
        devs = list()
        for path in glob.glob(SAS_DEVICE_FNMATCH):
            match = re.match(
                r'.*target\d+:\d+:\d+/(\d+):(\d+):(\d+):(\d+)', path)
            if not match:
                raise BlockProviderException(
                    BlockProviderException.format_error_message(
                        "No match for SAS Target {}".format(path)))
            adapter = match.group(1)
            channel = match.group(2)
            target_id = match.group(3)
            lun = match.group(4)
            devs.append(
                (adapter, channel, target_id, lun, os.path.basename(path)))
        return devs

    @staticmethod
    def rescan_hbas():
        """
        Trigger a device rescan of the HBAs
        """
        for path in glob.glob(FC_HOST_FNMATCH):
            with open(path, 'w') as rescan:
                rescan.write('- - -\n')

        # TODO: SAS?

        time.sleep(2)

    @staticmethod
    def get_paths(config, rescan=False):
        """
        Get the device paths for the block provider
        """
        devs = list()

        if rescan:
            HBA.rescan_hbas()

        fc_targets = HBA._get_fibre_channel_targets()
        devs.extend(HBA._get_fibre_channel_devices(fc_targets))

        devs.extend(HBA.get_sas_devices())

        return [HBA.make_blk_obj(adapter, channel, target_id, lun, dev)
                for (adapter, channel, target_id, lun, dev) in devs]


class HBABlockObj(BlockObj):
    """
    Private implementation class to add to BlockObj
    """

    def __init__(self, path, adapter, channel, target_id, lun, provider):
        super(HBABlockObj, self).__init__(path)
        self._provider = provider
        self._adapter = adapter
        self._channel = channel
        self._target_id = target_id
        self._lun = lun

    @property
    def provider(self):
        return self._provider

    def get_device_info(self):
        info = super(HBABlockObj, self).get_device_info()

        info['adapter'] = self._adapter
        info['channel'] = self._channel
        info['id'] = self._target_id
        info['lun'] = self._lun

        return info
