"""
Implementation of BlockProvider to handle iSCSI targets
"""

import json
import os  # temporary
import re
import uuid as uuid_mod  # it would be better to use a common one from util.py
from .fake import libiscsi
from .fake import util
from .fake import f_exceptions
from .blockprovider import BlockProvider
from .blockprovider import BlockList
from .blockprovider import BlockObj
from .blockprovider import BlockProviderException
from . import log


class ISCSIException(BlockProviderException):
    """Custom Exception class"""
    pass


class ISCSI(BlockProvider):
    """
    Class providing paths to iscsi devices.

    This is supposed to be a static class.
    We deal with only one IP and one IQN for the time being.
    """

    login_wait_to = 20

    @staticmethod
    def provider():
        """
        Return the provider key for this block provider
        """
        return "iscsi"

    def __init__(self, uuid, targets=None, username="", password="",
                 persistent=False, rescan=False):
        """
        targets: a list of ip/iqn pairs
        username: CHAP username for login
        password: CHAP password for login
        """
        self.uuid = uuid
        if targets is None:
            self.targets = []
        else:
            self.targets = targets
        self.logged_targets = set()
        self.username = username
        self.password = password
        self.persistent = persistent
        self.rescan = rescan
        self.path_list = []

    @staticmethod
    def _login(portal, target, username, password, uuid, rescan=False):
        """use libiscsi and let it raise whatever exception it raises"""

        log.sm_log("Logging into {}/{} (username={})".
                   format(portal, target, username))
        libiscsi.login(portal, target, username, password, uuid, rescan)

        # Some devices could be slow to appear and checking with udev settle
        # is not enough because a login causes many events (host creation, session,
        # target and, eventually, devices) and udev settle could return as soon as
        # the first events are processed but before the interesting ones (devices)
        # still have to be notified.
        # Our xs-mpath-scsidev.sh in udev scripts creates devices in /dev/iscsi and
        # this is what we could be waiting for.
        # Worst case: there are no devices and we wasted login_wait_to secs.
        # Worst worst case: the device is very slow and we think there are no devices but,
        # actually, we cannot do anything good with such a slow device.
        iscsi_path = '/dev/iscsi/'+target+'/'+portal+'/'
        if not util.wait_for_path(iscsi_path, ISCSI.login_wait_to):
            raise ISCSIException(ISCSIException.format_error_message("No mapped LUNs"))
        util.pread2(['udevadm', 'settle'])

        try:
            lun_list = os.listdir(iscsi_path)
        except OSError as exc:
            if exc.errno == 2:
                # If we are here, there was a race condition between login/logout
                raise ISCSIException(ISCSIException.format_error_message("No mapped LUNs"))
            else:
                raise

        # filter entries by name
        path_list = [(iscsi_path+entry, target, portal) for entry in lun_list
                     if entry.startswith('LUN')]

        return path_list

    def login(self):
        """Login to all the targets store in the object

        This function reads the list of ip/iqn pairs stored in self.targets,
        attempts to login to all of them, keeps track of the successful ones
        in self.logged_targets and stores the list of available device paths
        created by the login, into self.path_list

        Raises f_exceptions.XenError"""

        last_exc = None
        success = False

        for portal, tgt_iqn in self.targets:
            try:
                path_list = self._login(portal, tgt_iqn, self.username,
                                        self.password, self.uuid, self.rescan)
                self.path_list.extend(path_list)
                self.logged_targets.add((portal, tgt_iqn))
                success = True
            except (f_exceptions.XenError, ISCSIException) as exc:
                log.sm_log("Login failed for {}/{}: {}".
                           format(portal, tgt_iqn, exc))
                last_exc = exc

        if last_exc and not success:
            raise last_exc

    def logout(self):
        """Release all the iscsi sessions associated with this object

        A More appropriate name would be 'release' because this function will
        actually logout only if there are no other ISCSI objects with a different
        uuid using the iscsi session.

        A logout will be attempted on all the open sessions for this object but
        no more than once for every target IQN.
        This is because of the way low level refcounting of iscsi sessions works:
        it does not look at both ip and iqn, but only at the iqn and the uuid.
        This forces us to call logout on all the ips associated with that IQN
        at once. It won't interfere with other ISCSI objects because of the uuid.
        """

        iqn_done = set()
        targets_logout = set()  # elements to remove from self.logged_targets
        for ip, iqn in self.targets:
            if iqn in iqn_done:
                # This pair can be considered done
                targets_logout.add((ip, iqn))
                continue
            try:
                libiscsi.logout(iqn, self.uuid)
                iqn_done.add(iqn)
                targets_logout.add((ip, iqn))
            except f_exceptions.XenError as exc:
                log.sm_log("Logout failed for {}\n {}".format(iqn, exc.message))
        # If everything went fine, the resulting set must be empty
        self.logged_targets -= targets_logout
        if self.logged_targets:
            log.sm_log("After logout the following pairs are still present: {}".
                       format(self.logged_targets))

    def persist(self):
        """
        Set the connection persistent
        """
        self.persistent = True

    @staticmethod
    def discovery(target_ip, target_port, chapuser=None, chappass=None):
        """use libiscsi and let it raise whatever exception it raises"""

        if chapuser is None:
            chapuser, chappass = '', ''
        iqn_list = libiscsi.discovery(target_ip, target_port, chapuser,
                                      chappass)
        return iqn_list

    @staticmethod
    def _no_duplicate_targets(targets, tgt_list, known_tgt):
        """Remove duplicate entries from discovery results

        Discovery queries to different ips can return same results.

        This function is generic enough to remove duplicate entries
        from any group of iterables

        Args:
            targets: iterable to search for non existing entries
            tgt_list: list to pass that is filled with non repeated
               entries
            known_tgt: set containing existing entries in string form
        """
        for tgt in targets:
            key = ",".join(tgt)
            if key in known_tgt:
                continue
            known_tgt.add(key)
            tgt_list.append(tgt)

    @staticmethod
    def _parse_node_output(tgt_list, tgt_iqn):
        """Replacement of original parse_node_output.

        This function works on the results discovery function and
        not the command line.
        In the former case we have a list of tuples while in the
        latter the textual output of "iscsiadm discovery".

        It does not do de-duplication.
        """

        new_tgt_list = []
        if tgt_iqn == "*":
            new_tgt_list = tgt_list[:]
        else:
            new_tgt_list = [tgt for tgt in tgt_list if tgt[2] == tgt_iqn]
        return new_tgt_list

    @staticmethod
    def _create_iscsi_from_config(config, target_list, rescan=False):
        """Create an ISCSI instance from config

        config: the dictionary coming from dev_config
        target_list: list of targets as returned by ISCSI.discovery
        """

        target_iqn = config['iqns']

        # Get credentials if any
        username = config.get('chapuser', '')
        password = config.get('chappassword', '')

        iscsi_uuid = (
            config['uuid'] if 'uuid' in config
            else str(uuid_mod.uuid4()))
        target_list = ISCSI._parse_node_output(target_list, target_iqn)
        # remove the target portal group until we find good use of it
        target_list = [(x, y) for x, _, y in target_list]

        return ISCSI(iscsi_uuid, targets=target_list, username=username,
                     password=password, rescan=rescan)


    @staticmethod
    def _parse_options(config, rescan):
        """This function contains the logic needed to process ISCSI
        parameters.
        """

        if 'ips' not in config:
            raise ISCSIException(ISCSIException.format_error_message('Missing portal address'))

        # "ips" is a comma separated list of discovery targets.
        # For the time being, no port and no DNS resolution.
        discovery_ip_list = config['ips'].split(',')

        known_targets = set()
        target_list = []
        for discovery_ip in discovery_ip_list:
            targets = ISCSI.discovery(discovery_ip, '3260')
            ISCSI._no_duplicate_targets(targets, target_list, known_targets)

        # For the time being, we accept either one iqn or all. Same as old API.
        # This limitation is not necessary and we can easily accept a comma
        # separated list.
        if 'iqns' in config:
            iscsi_obj = ISCSI._create_iscsi_from_config(
                config, target_list, rescan)
            iscsi_obj.login()
            blk_list = BlockList()
            for path, iqn, portal in iscsi_obj.path_list:
                blk_obj = ISCSI._make_blk_obj(path, portal, iqn, iscsi_obj)
                blk_list.add(blk_obj)

            return blk_list
        else:
            return target_list

    @staticmethod
    def get_paths(config, rescan=False):
        """
        Get the device paths for the block provider
        """
        output = ISCSI._parse_options(config, rescan)
        if not isinstance(output, BlockList):
            iqn_list = []
            for entry in output:
                iqn_list.append(entry)
            raise ISCSIException(
                ISCSIException.format_json_options(
                    'iqns',
                    [{'value': z, 'info': {}} for (_, _, z) in iqn_list]))
        return output

    @classmethod
    def _make_blk_obj(cls, path, portal, iqn, iscsi_obj):
        return IscsiBlockObj(path, portal, iqn, iscsi_obj)

    def __str__(self):
        return ("ISCSI obj {}\n"
                "  targets={}\n"
                "  logged={}".format(self.uuid,
                                     self.targets,
                                     self.logged_targets))

    def __del__(self):
        if not self.persistent and self.logged_targets:
            log.sm_log("No more use for ISCSI obj {}\n"
                       "  Logging out all open sessions".format(
                           self.uuid))
            self.logout()


class IscsiBlockObj(BlockObj):
    """ ISCSI BlockObj """

    def __init__(self, path, portal, tgt_iqn, iscsi_obj):
        super(IscsiBlockObj, self).__init__(path)
        self.portal = portal
        self.tgt_iqn = tgt_iqn
        self.iscsi_obj = iscsi_obj  # ISCSI obj in charge for this device
        self._provider = iscsi_obj.provider()

    @property
    def provider(self):
        return self._provider

    def close(self):
        self.iscsi_obj.logout()

    def persist(self):
        """
        Set the connection to persistent
        """
        log.sm_log("Setting iscsi to persistent {}".format(
            self.iscsi_obj.uuid))
        super(IscsiBlockObj, self).persist()
        self.iscsi_obj.persist()

    def get_device_info(self):
        info = super(IscsiBlockObj, self).get_device_info()

        match = re.match(r'^.*/LUN(\d+)$', self._path)
        if match:
            info['LUNid'] = match.group(1)

        return info
