"""
Mapper for taking local kernel devices and converting to SCSI IDs
Better rename it later to something like 'WWID'
"""

import os

from .fake import util, f_exceptions
from .fake import scsiutil
from . import log


class ScsiIdObj(object):
    """Object mapping device paths to scsi_id"""

    def __init__(self, scsi_id, path):
        self._scsi_id = scsi_id
        self._path = path
        self._size = 0
        self._devs = []

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def add(self, dev):
        """add a DevObj to the list of devices with same scsi id"""
        self._devs.append(dev)

    def close(self):
        log.sm_log("Closing {}".format(self.scsi_id))

        for dev in self._devs:
            dev.close()

    @property
    def path(self):
        """ Access the device path for this scsi object """
        return self._path

    @property
    def scsi_id(self):
        return self._scsi_id

    @property
    def size(self):
        return scsiutil.getsize(self.path)

    @property
    def vendor(self):
        return scsiutil.get_vendor(self.path)

    @property
    def serial(self):
        return scsiutil.get_serial(self.path)

    def set_persistent(self):
        """Notify the lower layers that the block device(s)
        used for this object needs to be persistent.
        """

        # _devs cannot be empty: make sure this is enforced
        for dev in self._devs:
            dev.persist()

    def rescan(self):
        scsiutil.refresh_lun_size_by_SCSIid(self.scsi_id)

    def get_device_info(self):
        info = {}

        info['vendor'] = self.vendor
        info['serial'] = self.serial

        for dev in self._devs:
            info.update(dev.get_device_info())

        return info

    def __str__(self):
        """
        Get a human useful string
        """
        return '{} => {}'.format(self.path, self.serial)


class ScsiIdList(object):
    """Object containing a list of ScsiIdObj"""

    def __init__(self):
        self._objs = {}

    def __iter__(self):
        return self._objs.itervalues()

    def add(self, scsi_id, scsi_obj):
        """Add a ScsiIdObj to the list"""
        self._objs[scsi_id] = scsi_obj

    def update(self, scsi_id, dev):
        """Add a DevObj to the existing ScsiIdObj with the same id"""
        self._objs[scsi_id].add(dev)

    def get_obj(self, scsi_id):
        """Get the corresponding ScsiIdObj"""
        return self._objs[scsi_id] if scsi_id in self._objs else None


class ScsiIdMapperException(Exception):
    """
    General exception type for ScsiIdMapper errors

    Message should be formatted JSON response.
    """
    pass


class ScsiIdMapper(object):
    """
    Mapper class to convert BlockProvider to a set of SCSI IDs
    """

    def __init__(self):
        raise ScsiIdMapperException("Cannot create an instance")

    @staticmethod
    def get_devs(dev_list):
        """
        Get a list of devices, add their scsi_id and return a list of objects
        organized by scsi_id
        """

        path_pref = '/dev/disk/by-id/scsi-'

        scsi_id_list = ScsiIdList()
        for obj in dev_list:
            try:
                device_path = obj.get_path()
                scsi_id = util.get_scsi_id(device_path)
                obj.set_serial(scsi_id)
                # The following uses explicitly _objs since we changed
                # the __iter__ method in ScsiIdList
                if scsi_id in scsi_id_list._objs:
                    scsi_id_list.update(scsi_id, obj)
                    continue
                scsi_id_path = path_pref + scsi_id
                # Check that the scsi-XXXXX path exists (it doesn't for SAS)
                if os.path.exists(scsi_id_path):
                    device_path = scsi_id_path
                scsi_obj = ScsiIdObj(scsi_id, device_path)
                scsi_obj.add(obj)
                scsi_id_list.add(scsi_id, scsi_obj)
            except f_exceptions.XenError:
                # Ignore devices that fail to give an id
                log.sm_log('Failed to get SCSI id for {}, ignoring'.format(
                    obj.get_path()))
                # TODO: Ask the device object to remove from kernel?
        return scsi_id_list
